@@ -33,7 +33,6 @@
 
 /****************************************************************************
 **
-
 *F  IS_INTOBJ( <o> )  . . . . . . . .  test if an object is an integer object
 **
 **  'IS_INTOBJ' returns 1 if the object <o> is an (immediate) integer object,
@@ -43,6 +42,16 @@
     ((libGAP_Int)(o) & 0x01)
 
 
+/****************************************************************************
+**
+*F  IS_POS_INTOBJ( <o> )  . .  test if an object is a positive integer object
+**
+**  'IS_POS_INTOBJ' returns 1 if the object <o> is an (immediate) integer
+**  object encoding a positive integer, and 0 otherwise.
+*/
+#define libGAP_IS_POS_INTOBJ(o) \
+    (((libGAP_Int)(o) & 0x01) && ((libGAP_Int)(o) > 0x01))
+
 
 /****************************************************************************
 **
@@ -137,11 +146,11 @@
 */
 #if HAVE_ARITHRIGHTSHIFT
 #define libGAP_DIFF_INTOBJS(o,l,r)            \
-    ((o) = (libGAP_Bag)((libGAP_Int)(l)-(libGAP_Int)(r)+1), \
+    ((o) = (libGAP_Obj)((libGAP_Int)(l)-(libGAP_Int)(r)+1), \
      (((libGAP_Int)(o) << 1) >> 1) == (libGAP_Int)(o) )
 #else
 #define libGAP_DIFF_INTOBJS(o,l,r)            \
-    ((o) = (libGAP_Bag)((libGAP_Int)(l)-(libGAP_Int)(r)+1), \
+    ((o) = (libGAP_Obj)((libGAP_Int)(l)-(libGAP_Int)(r)+1), \
      ((((libGAP_UInt) (o)) >> (sizeof(libGAP_UInt)*8-2))-1) > 1)
 #endif
 
@@ -161,7 +170,6 @@
 #define libGAP_HALF_A_WORD 16
 #endif
 
-#if HAVE_ARITHRIGHTSHIFT
 static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
 {
   libGAP_Int prod;
@@ -172,39 +180,30 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
   if (r == (libGAP_Int)libGAP_INTOBJ_INT(1))
     return (libGAP_Obj)l;
   prod = ((libGAP_Int)((libGAP_UInt)l >> 2) * ((libGAP_UInt)r-1)+1);
+
+#if HAVE_ARITHRIGHTSHIFT
   if ((prod << 1)>> 1 !=  prod)
     return (libGAP_Obj) 0;
-  if ((((libGAP_Int)l)<<libGAP_HALF_A_WORD)>>libGAP_HALF_A_WORD == (libGAP_Int) l &&
-      (((libGAP_Int)r)<<libGAP_HALF_A_WORD)>>libGAP_HALF_A_WORD == (libGAP_Int) r)
-    return (libGAP_Obj) prod;
-  if ((prod -1) / (l >> 2) == r-1)
-    return (libGAP_Obj) prod;
-  else
-    return (libGAP_Obj) 0;
-}
 #else
-static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
-{
-  libGAP_Int prod;
-  if (l == (libGAP_Int)libGAP_INTOBJ_INT(0) || r == (libGAP_Int)libGAP_INTOBJ_INT(0))
-    return libGAP_INTOBJ_INT(0);
-  if (l == (libGAP_Int)libGAP_INTOBJ_INT(1))
-    return (libGAP_Obj)r;
-  if (r == (libGAP_Int)libGAP_INTOBJ_INT(1))
-    return (libGAP_Obj)l;
-  prod = ((libGAP_Int)((libGAP_UInt)l >> 2) * ((libGAP_UInt)r-1)+1);
   if (((((libGAP_UInt) (prod)) >> (sizeof(libGAP_UInt)*8-2))-1) <= 1)
     return (libGAP_Obj) 0;
+#endif
+
   if ((((libGAP_Int)l)<<libGAP_HALF_A_WORD)>>libGAP_HALF_A_WORD == (libGAP_Int) l &&
       (((libGAP_Int)r)<<libGAP_HALF_A_WORD)>>libGAP_HALF_A_WORD == (libGAP_Int) r)
     return (libGAP_Obj) prod;
+
+#if HAVE_ARITHRIGHTSHIFT
+  if ((prod -1) / (l >> 2) == r-1)
+    return (libGAP_Obj) prod;
+#else
   if ((prod-1) / ((l-1)/4) == r-1)
     return (libGAP_Obj) prod;
-  else
-    return (libGAP_Obj) 0;
-}
 #endif
 
+  return (libGAP_Obj) 0;
+}
+
 #define libGAP_PROD_INTOBJS( o, l, r) ((o) = libGAP_prod_intobjs((libGAP_Int)(l),(libGAP_Int)(r)), \
                                   (o) != (libGAP_Obj) 0)
    
@@ -219,6 +218,20 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
                         ((libGAP_Int)(o) & 0x02)
 
 
+/****************************************************************************
+**
+*F  RegisterPackageTNUM( <name>, <typeObjFunc> )
+**
+**  Allocates a TNUM for use by a package. The parameters <name> and
+**  <typeObjFunc> are used to initialize the relevant entries in the
+**  InfoBags and TypeObjFuncs arrays.
+**
+**  If allocation fails (e.g. because no more TNUMs are available),
+**  a negative value is returned.
+*/
+libGAP_Int libGAP_RegisterPackageTNUM( const char *name, libGAP_Obj (*typeObjFunc)(libGAP_Obj obj) );
+
+
 /****************************************************************************
 **
 
@@ -263,7 +276,7 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
 */
 #define libGAP_FIRST_REAL_TNUM         0
 
-#define libGAP_FIRST_CONSTANT_TNUM     ((libGAP_UInt)0)
+#define libGAP_FIRST_CONSTANT_TNUM     (0UL)
 #define libGAP_T_INT                   (libGAP_FIRST_CONSTANT_TNUM+ 0)    /* immediate */
 #define libGAP_T_INTPOS                (libGAP_FIRST_CONSTANT_TNUM+ 1)
 #define libGAP_T_INTNEG                (libGAP_FIRST_CONSTANT_TNUM+ 2)
@@ -339,7 +352,12 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
 #define libGAP_T_WPOBJ                 (libGAP_FIRST_EXTERNAL_TNUM+ 3)
      /* #define T_DUMMYOBJ              (FIRST_EXTERNAL_TNUM+ 4)
         remove to get parity right */
-#define libGAP_LAST_EXTERNAL_TNUM      libGAP_T_WPOBJ
+
+/* reserve space for 50 package TNUMs */
+#define libGAP_FIRST_PACKAGE_TNUM      (libGAP_FIRST_EXTERNAL_TNUM+ 4)
+#define libGAP_LAST_PACKAGE_TNUM       (libGAP_FIRST_EXTERNAL_TNUM+53)
+
+#define libGAP_LAST_EXTERNAL_TNUM      libGAP_LAST_PACKAGE_TNUM
 #define libGAP_LAST_REAL_TNUM          libGAP_LAST_EXTERNAL_TNUM
 #define libGAP_LAST_VIRTUAL_TNUM libGAP_LAST_EXTERNAL_TNUM
 
@@ -352,66 +370,26 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
 #define libGAP_TESTING                 libGAP_COPYING
 #define libGAP_LAST_TESTING_TNUM       libGAP_LAST_COPYING_TNUM
 
+#if libGAP_LAST_COPYING_TNUM > 254
+#error libGAP_LAST_COPYING_TNUM out of range
+#endif
 
 
 /****************************************************************************
 **
-
-*F  F_MUTABLE . . . . . . . . . . . . . . . . . . . . . . .  IsMutableObjFilt
+*S  T_BODY  . . . . . . . . . . . . . . . . . . . . type of function body bag
+**
+**  'T_BODY' is the type of the function body bags.
 */
-#define libGAP_F_MUTABLE       1
+#define libGAP_T_BODY                  254
+
+#if libGAP_T_BODY <= libGAP_LAST_COPYING_TNUM
+#error libGAP_T_BODY out of range
+#endif
 
 
 /****************************************************************************
 **
-*F  F_EMPTY . . . . . . . . . . . . . . . . . . . . . . . . . . . IsEmptyProp
-*/
-#define libGAP_F_EMPTY         2
-#define libGAP_F_NOT_EMPTY     3
-
-
-/****************************************************************************
-**
-*F  F_SSORT . . . . . . . . . . . . . . . . . . . . . . . . . . . IsSSortProp
-*/
-#define libGAP_F_SSORT         4
-#define libGAP_F_NOT_SSORT     5
-
-
-/****************************************************************************
-**
-*F  F_DENSE . . . . . . . . . . . . . . . . . . . . . . . . . . . IsDenseProp
-*/
-#define libGAP_F_DENSE         6
-#define libGAP_F_NOT_DENSE     7
-
-
-/****************************************************************************
-**
-*F  F_HOMOG . . . . . . . . . . . . . . . . . . . . . . . . . . . IsHomogProp
-*/
-#define libGAP_F_HOMOG         8
-#define libGAP_F_NOT_HOMOG     9
-
-
-/****************************************************************************
-**
-*F  F_TABLE . . . . . . . . . . . . . . . . . . . . . . . . . . . IsTableProp
-*/
-#define libGAP_F_TABLE         10
-#define libGAP_F_NOT_TABLE     11
-
-/****************************************************************************
-**
-*F  F_RECT . . . . . . . . . . . . . . . . . . . . . . IsRectangularTableProp
-*/
-#define libGAP_F_RECT         12
-#define libGAP_F_NOT_RECT     13
-
-
-/****************************************************************************
-**
-
 *F  TNUM_OBJ( <obj> ) . . . . . . . . . . . . . . . . . . . type of an object
 **
 **  'TNUM_OBJ' returns the type of the object <obj>.
@@ -449,11 +427,11 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
 /****************************************************************************
 **
 
-*F  FAMILY_TYPE( <kind> ) . . . . . . . . . . . . . . . . .  family of a kind
+*F  FAMILY_TYPE( <type> ) . . . . . . . . . . . . . . . . .  family of a type
 **
-**  'FAMILY_TYPE' returns the family of the kind <kind>.
+**  'FAMILY_TYPE' returns the family of the type <type>.
 */
-#define libGAP_FAMILY_TYPE(kind)       libGAP_ELM_PLIST( kind, 1 )
+#define libGAP_FAMILY_TYPE(type)       libGAP_ELM_PLIST( type, 1 )
 
 
 /****************************************************************************
@@ -465,44 +443,52 @@ static inline libGAP_Obj libGAP_prod_intobjs(libGAP_Int l, libGAP_Int r)
 
 /****************************************************************************
 **
-*F  FLAGS_TYPE( <kind> )  . . . . . . . . . . .  flags boolean list of a kind
+*F  FLAGS_TYPE( <type> )  . . . . . . . . . . .  flags boolean list of a type
 **
-**  'FLAGS_TYPE' returns the flags boolean list of the kind <kind>.
+**  'FLAGS_TYPE' returns the flags boolean list of the type <type>.
 */
-#define libGAP_FLAGS_TYPE(kind)        libGAP_ELM_PLIST( kind, 2 )
+#define libGAP_FLAGS_TYPE(type)        libGAP_ELM_PLIST( type, 2 )
 
 
 /****************************************************************************
 **
-*F  SHARED_TYPE( <kind> ) . . . . . . . . . . . . . . . shared data of a kind
+*F  SHARED_TYPE( <type> ) . . . . . . . . . . . . . . . shared data of a type
 **
-**  'SHARED_TYPE' returns the shared data of the kind <kind>.
+**  'SHARED_TYPE' returns the shared data of the type <type>.
 XXX nowhere used, throw away??? (FL)
 */
-/* #define SHARED_TYPE(kind)       ELM_PLIST( kind, 3 )
+/* #define SHARED_TYPE(type)       ELM_PLIST( type, 3 )
 */                        
                         
 /****************************************************************************
 **
-*F  ID_TYPE( <kind> ) . . . . . . . . . . . . . . . . . . . . .  id of a kind
+*F  ID_TYPE( <type> ) . . . . . . . . . . . . . . . . . . . . .  id of a type
 **
-**  'ID_TYPE' returns the ID of  a kind.  Warning: if  GAP runs out of ID  it
+**  'ID_TYPE' returns the ID of  a type.  Warning: if  GAP runs out of ID  it
 **  will renumber all IDs.  Therefore the  corresponding routine must excatly
 **  know where such numbers are stored.
 */
-#define libGAP_ID_TYPE(kind)           libGAP_ELM_PLIST( kind, 4 )
+#define libGAP_ID_TYPE(type)           libGAP_ELM_PLIST( type, 4 )
 
 
 /****************************************************************************
 **
-*F  TYPE_OBJ( <obj> ) . . . . . . . . . . . . . . . . . . . kind of an object
+*F  TYPE_OBJ( <obj> ) . . . . . . . . . . . . . . . . . . . type of an object
 **
-**  'TYPE_OBJ' returns the kind of the object <obj>.
+**  'TYPE_OBJ' returns the type of the object <obj>.
 */
 #define libGAP_TYPE_OBJ(obj)   ((*libGAP_TypeObjFuncs[ libGAP_TNUM_OBJ(obj) ])( obj ))
 
-extern libGAP_Obj (*libGAP_TypeObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) ( libGAP_Obj obj );
+extern libGAP_Obj (*libGAP_TypeObjFuncs[libGAP_LAST_REAL_TNUM+1]) ( libGAP_Obj obj );
 
+/****************************************************************************
+**
+*F  SetTypeDatobj( <obj>, <kind> ) . . . . . . . .  set kind of a data object
+**
+**  'SetTypeDatobj' sets the kind <kind> of the data object <obj>.
+*/
+
+#define libGAP_SetTypeDatObj(obj, type)  libGAP_SET_TYPE_DATOBJ(obj, type)
 
 /****************************************************************************
 **
@@ -540,7 +526,7 @@ extern void libGAP_MakeImmutable( libGAP_Obj obj );
 #define libGAP_IS_MUTABLE_OBJ(obj) \
                         ((*libGAP_IsMutableObjFuncs[ libGAP_TNUM_OBJ(obj) ])( obj ))
 
-extern libGAP_Int (*libGAP_IsMutableObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) ( libGAP_Obj obj );
+extern libGAP_Int (*libGAP_IsMutableObjFuncs[libGAP_LAST_REAL_TNUM+1]) ( libGAP_Obj obj );
 
 /****************************************************************************
 **
@@ -557,9 +543,9 @@ extern libGAP_Int (*libGAP_IsMutableObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) ( libGA
 **  No saving function may allocate any bag
 */
 
-extern void (*libGAP_SaveObjFuncs[ 256 ]) (libGAP_Obj obj);
+extern void (*libGAP_SaveObjFuncs[256]) ( libGAP_Obj obj );
 
-extern void libGAP_SaveObjError ( libGAP_Obj obj );
+extern void libGAP_SaveObjError( libGAP_Obj obj );
 
 
 /****************************************************************************
@@ -577,11 +563,9 @@ extern void libGAP_SaveObjError ( libGAP_Obj obj );
 **  No loading function may allocate any bag
 */
 
-extern void (*libGAP_LoadObjFuncs[ 256 ]) (libGAP_Bag bag);
+extern void (*libGAP_LoadObjFuncs[256]) ( libGAP_Obj obj );
 
-extern void libGAP_LoadObjError (
-                   libGAP_Bag bag
-                   );
+extern void libGAP_LoadObjError( libGAP_Obj obj );
 
 /****************************************************************************
 **
@@ -593,7 +577,7 @@ extern void libGAP_LoadObjError (
 #define libGAP_IS_COPYABLE_OBJ(obj) \
                         ((libGAP_IsCopyableObjFuncs[ libGAP_TNUM_OBJ(obj) ])( obj ))
 
-extern libGAP_Int (*libGAP_IsCopyableObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) ( libGAP_Obj obj );
+extern libGAP_Int (*libGAP_IsCopyableObjFuncs[libGAP_LAST_REAL_TNUM+1]) ( libGAP_Obj obj );
 
 
 /****************************************************************************
@@ -611,7 +595,7 @@ extern libGAP_Int (*libGAP_IsCopyableObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) ( libG
 **
 *V  ShallowCopyObjFuncs[<type>] . . . . . . . . . .  shallow copier functions
 */
-extern libGAP_Obj (*libGAP_ShallowCopyObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) ( libGAP_Obj obj );
+extern libGAP_Obj (*libGAP_ShallowCopyObjFuncs[libGAP_LAST_REAL_TNUM+1]) ( libGAP_Obj obj );
 
 
 /****************************************************************************
@@ -676,7 +660,7 @@ extern libGAP_Obj libGAP_CopyObj (
 **  then call 'CLEAN_OBJ'  for all subobjects recursively.  If called for an
 **  already unmarked object, it should simply return.
 */
-extern libGAP_Obj (*libGAP_CopyObjFuncs[ libGAP_LAST_REAL_TNUM+libGAP_COPYING+1 ]) ( libGAP_Obj obj, libGAP_Int mut );
+extern libGAP_Obj (*libGAP_CopyObjFuncs[libGAP_LAST_REAL_TNUM+libGAP_COPYING+1]) ( libGAP_Obj obj, libGAP_Int mut );
 
 
 
@@ -684,10 +668,10 @@ extern libGAP_Obj (*libGAP_CopyObjFuncs[ libGAP_LAST_REAL_TNUM+libGAP_COPYING+1
 **
 *V  CleanObjFuncs[<type>] . . . . . . . . . . . . table of cleaning functions
 */
-extern void (*libGAP_CleanObjFuncs[ libGAP_LAST_REAL_TNUM+libGAP_COPYING+1 ]) ( libGAP_Obj obj );
+extern void (*libGAP_CleanObjFuncs[libGAP_LAST_REAL_TNUM+libGAP_COPYING+1]) ( libGAP_Obj obj );
 
 
-extern void (*libGAP_MakeImmutableObjFuncs[ libGAP_LAST_REAL_TNUM+1 ]) (libGAP_Obj obj );
+extern void (*libGAP_MakeImmutableObjFuncs[libGAP_LAST_REAL_TNUM+1]) ( libGAP_Obj obj );
 
 /****************************************************************************
 **
@@ -715,7 +699,7 @@ extern libGAP_Int  libGAP_PrintObjDepth;
 
 extern libGAP_Int  libGAP_PrintObjFull;
 
-extern void (* libGAP_PrintObjFuncs [ libGAP_LAST_REAL_TNUM  +1 ]) ( libGAP_Obj obj );
+extern void (* libGAP_PrintObjFuncs[libGAP_LAST_REAL_TNUM+1]) ( libGAP_Obj obj );
 
 
 /****************************************************************************
@@ -742,7 +726,7 @@ extern void libGAP_ViewObj (
 **
 **  These are also used for viewing
 */
-extern void (* libGAP_PrintPathFuncs [ libGAP_LAST_REAL_TNUM  +1 ]) (
+extern void (* libGAP_PrintPathFuncs[libGAP_LAST_REAL_TNUM+1]) (
     libGAP_Obj                 obj,
     libGAP_Int                 indx );
 
@@ -757,14 +741,14 @@ extern void (* libGAP_PrintPathFuncs [ libGAP_LAST_REAL_TNUM  +1 ]) (
 
 /****************************************************************************
 **
-*F  TYPE_COMOBJ( <obj> )  . . . . . . . . . . . .  kind of a component object
+*F  TYPE_COMOBJ( <obj> )  . . . . . . . . . . . .  type of a component object
 */
 #define libGAP_TYPE_COMOBJ(obj)          libGAP_ADDR_OBJ(obj)[0]
 
 
 /****************************************************************************
 **
-*F  SET_TYPE_COMOBJ( <obj>, <val> ) . . .  set the kind of a component object
+*F  SET_TYPE_COMOBJ( <obj>, <val> ) . . .  set the type of a component object
 */
 #define libGAP_SET_TYPE_COMOBJ(obj,val)  (libGAP_ADDR_OBJ(obj)[0] = (val))
 
@@ -779,14 +763,14 @@ extern void (* libGAP_PrintPathFuncs [ libGAP_LAST_REAL_TNUM  +1 ]) (
 
 /****************************************************************************
 **
-*F  TYPE_POSOBJ( <obj> )  . . . . . . . . . . . . kind of a positional object
+*F  TYPE_POSOBJ( <obj> )  . . . . . . . . . . . . type of a positional object
 */
 #define libGAP_TYPE_POSOBJ(obj)          libGAP_ADDR_OBJ(obj)[0]
 
 
 /****************************************************************************
 **
-*F  SET_TYPE_POSOBJ( <obj>, <val> ) . . . set the kind of a positional object
+*F  SET_TYPE_POSOBJ( <obj>, <val> ) . . . set the type of a positional object
 */
 #define libGAP_SET_TYPE_POSOBJ(obj,val)  (libGAP_ADDR_OBJ(obj)[0] = (val))
  
@@ -801,25 +785,14 @@ extern void (* libGAP_PrintPathFuncs [ libGAP_LAST_REAL_TNUM  +1 ]) (
 
 /****************************************************************************
 **
-*F  TYPE_DATOBJ( <obj> )  . . . . . . . . . . . . . . . kind of a data object
+*F  TYPE_DATOBJ( <obj> )  . . . . . . . . . . . . . . . type of a data object
 */
 #define libGAP_TYPE_DATOBJ(obj)          libGAP_ADDR_OBJ(obj)[0]
 
 
 /****************************************************************************
 **
-*F  TYPE_ANYOBJ( <obj> )  . . . . . . . . . . . . .kind of an external object
-**
-** This is a hack which relies on the fact that TYPE_COMOBJ, TYPE_DATOBJ and
-** TYPE_POSOBJ are actually the same
-*/
-
-#define libGAP_TYPE_ANYOBJ(obj)          libGAP_ADDR_OBJ(obj)[0]
-
-
-/****************************************************************************
-**
-*F  SET_TYPE_DATOBJ( <obj>, <val> )  . . . . .  set the kind of a data object
+*F  SET_TYPE_DATOBJ( <obj>, <val> )  . . . . .  set the type of a data object
 */
 #define libGAP_SET_TYPE_DATOBJ(obj,val)  (libGAP_ADDR_OBJ(obj)[0] = (val))
 
